% SDPHapInfer v0.4, revised by YT at 20040826, Include and modify HAPAR's algorithm to reduce the 2s.
% SDPHapInfer v0.3, revised by YT at 20040824, Include Gustfield's algorithm to reduce the 2s.
% SDPHapInfer v0.2, revised by YT at 20040709, create a ResolGeno2 array to fix bugs.
% SDPHapInfer v0.1, created by YT at 20040630

%Program Flow:
%1. Read input file;
%2. Enumerate all possible haplotypes and build an index table;
%3. Build the constraint matrix for each genotype;
%   (slack variables are used, and the SDP problem is formulated in dual format)
%4. Build the objective and other constraint matrix;
%5. Call SDPA library to solve the SDP problem;
%6. Perform incomplete Choleskey factorization;
%7. Randomly selected two unit vectors for randomized rounding;
%8. Check if any genotype is still unresolved.

clear all;

fprintf('\n\n****************************************\n');
fprintf('*********** SDPHapInfer v0.4 ***********\n');
fprintf('****************************************\n');

FileName = input('Please specify the input file name: ', 's');
[NumOfGenotype, NumOfSNP, HapMat] = ReadGeno(FileName);
        
%Set minimum current number of haplotypes
CurrentHaplotype = 10000;

for iterations = 1:5
	fprintf('The %d run!\n',iterations);

	if (NumOfGenotype > 0)
		%Start timer
		tic;
	
		%Create the output file: xxx.out
		%OutputFileName = [FileName '.out'];
		OutputFileName = ['sdp.tmp'];			
		Outfid = fopen(OutputFileName, 'w');
		
		%Record the number of unresolved genotypes
		NumOfUGenotype = NumOfGenotype;
	
		%Create an array for the resolved genotype
		ResolGeno = zeros(1, NumOfGenotype);
		%Fix the bug in the version 0.1; ResolGeno2 array is not updated until the end of the while loop; 
		ResolGeno2 = zeros(1, NumOfGenotype);
		FinalSolution = [''];
	
		NumOfIteration = 0;
		
		%Start resolving each genotype until all of them are resolved
		while(NumOfUGenotype > 0)
			%Enumerate all possible haplotypes and build a MappingTable for indexing
			MappingTable = [''];
			tempMat = [''];
	
			%Record the number of variables required
			MaxIndex = 0;
	
			NumOfIteration = NumOfIteration + 1;
			
			fprintf('\n*******************************************************************\n');
			fprintf('*********** %d-th iteration of the SDPHapInfer algorithm ***********\n', NumOfIteration);
			fprintf('*******************************************************************\n');
			fprintf('There are %d genotypes still unresolved...\n\n', NumOfUGenotype);
			
			%Create a cell arrary to store the constraint matrix
			ConstMat = [];
			
			%Find the unresolved genotypes
			UResolGeno = find(ResolGeno == 0);
			
			NeedResolve = zeros(1,NumOfUGenotype);
			NeedSolution = ones(1,NumOfUGenotype);
			
			%Build the mapping table	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%% Gustfield and HAPAR's Speedup Algorithm %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
			GenoCell = cell(1, NumOfGenotype);
			
			for k1 = 1:NumOfUGenotype-1
				for k2 = k1+1:NumOfUGenotype
					Genotype1 = HapMat(UResolGeno(k1),:);
					Genotype2 = HapMat(UResolGeno(k2),:);
					if(strcmp(Genotype1,Genotype2) == 0)
					%if Genotype1 is not equal to Genotype2
						NumOf2 = 0;
						
						%Check each SNP
						Not2 = [];
						for l = 1:NumOfSNP
							if(Genotype1(l)==Genotype2(l))
								if(Genotype1(l) == '2')
									NumOf2 = NumOf2 + 1;
								else
									Not2 = [Not2,l];
								end
								continue;
							elseif(Genotype1(l)== '2' & Genotype2(l) ~= '2')
								Genotype1(l) = Genotype2(l);
								Not2 = [Not2,l];
							elseif(Genotype2(l)== '2' & Genotype1(l) ~= '2')
								Genotype2(l) = Genotype1(l);
								Not2 = [Not2,l];
							else
								%Two genotype are different
								NumOf2 = -1;
								break;
							end %end of else
						end %end of for loop l
						
						if(NumOf2 == -1)
							%this two genotypes are different
							%Gustfield's prevention method
							continue;
						elseif(NumOf2 == 0)
							%No 2s appear in the merged genotype
							p2 = Invert(Genotype1, HapMat(UResolGeno(k1),:));
							p3 = Invert(Genotype1, HapMat(UResolGeno(k2),:));
							%strcmp(Genotype1, p2) | strcmp(Genotype1, p3): 101, 202
							%Chk_Resolve to check if this haplotype can resolve other genotype
							if(NeedSolution(k1) | NeedSolution(k2) | Chk_Resolve(Genotype1, UResolGeno(k1), UResolGeno(k2), HapMat) | Chk_Resolve(p2, UResolGeno(k1), UResolGeno(k2), HapMat) | Chk_Resolve(p3, UResolGeno(k1), UResolGeno(k2), HapMat) | strcmp(Genotype1, p2) | strcmp(Genotype1, p3))
								NeedSolution(k1) = 0;
								NeedSolution(k2) = 0;
								NeedResolve(k1) = 1;
								NeedResolve(k2) = 1;
								
								if(~ismember(Genotype1, tempMat, 'rows'))
									tempMat = union(tempMat, Genotype1, 'rows');
									MaxIndex = MaxIndex + 1;
									MappingTable = [MappingTable; Genotype1];
									GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},MaxIndex); 
									GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},MaxIndex); 
								else
									tempIndex = Find_Index(Genotype1, MappingTable);
									GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},tempIndex); 
									GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},tempIndex); 
								end
								if(~ismember(p2, tempMat, 'rows'))
									tempMat = union(tempMat, p2, 'rows');
									MaxIndex = MaxIndex + 1;
									MappingTable = [MappingTable; p2];
									GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},MaxIndex); 
								else
									tempIndex = Find_Index(p2, MappingTable);
									GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},tempIndex); 
								end
								if(~ismember(p3, tempMat, 'rows'))
									tempMat = union(tempMat, p3, 'rows');
									MaxIndex = MaxIndex + 1;
									MappingTable = [MappingTable; p3];
									GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},MaxIndex); 
								else
									tempIndex = Find_Index(p3, MappingTable);
									GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},tempIndex); 
								end
							end
						else
							%There are 2s appearing in the merged genotype
							%Enumerate each possible haplotype
							for k3 = 0:2^(NumOf2)-1
								%create binary representation of one haplotype
								p1=dec2bin(k3);
								
								for m=1:NumOf2-length(p1)
									p1=['0', p1];
								end
								
								%Insert original 1 or 0 into the haplotypes
								for l = 1:length(Not2)
									if(Not2(l)==1)
										p1=[Genotype1(Not2(l)), p1];
									elseif(Not2(l) == NumOfSNP)
										p1=[p1, Genotype1(Not2(l))];
									else
										p1=[p1(1:(Not2(l)-1)), Genotype1(Not2(l)), p1(Not2(l):length(p1))];
									end %end of else
								end %end of for loop l
								
								p2 = Invert(p1, HapMat(UResolGeno(k1),:));
								p3 = Invert(p1, HapMat(UResolGeno(k2),:));
							
								if(NeedSolution(k1) | NeedSolution(k2) | Chk_Resolve(p1, UResolGeno(k1), UResolGeno(k2), HapMat) | Chk_Resolve(p2, UResolGeno(k1), UResolGeno(k2), HapMat) | Chk_Resolve(p3, UResolGeno(k1), UResolGeno(k2), HapMat))
									NeedSolution(k1) = 0;
									NeedSolution(k2) = 0;
									NeedResolve(k1) = 1;
									NeedResolve(k2) = 1;
									if(~ismember(p1, tempMat, 'rows'))
										tempMat = union(tempMat, p1, 'rows');
										MaxIndex = MaxIndex + 1;
										MappingTable = [MappingTable; p1];
										GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},MaxIndex); 
										GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},MaxIndex); 
									else
										tempIndex = Find_Index(p1, MappingTable);
										GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},tempIndex); 
										GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},tempIndex); 
									end
									if(~ismember(p2, tempMat, 'rows'))
										tempMat = union(tempMat, p2, 'rows');
										MaxIndex = MaxIndex + 1;
										MappingTable = [MappingTable; p2];
										GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)}, MaxIndex); 
									else
										tempIndex = Find_Index(p2, MappingTable);
										GenoCell{UResolGeno(k1)} = union(GenoCell{UResolGeno(k1)},tempIndex); 
									end
									if(~ismember(p3, tempMat, 'rows'))
										tempMat = union(tempMat, p3, 'rows');
										MaxIndex = MaxIndex + 1;
										MappingTable = [MappingTable; p3];
										GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)}, MaxIndex); 
									else
										tempIndex = Find_Index(p3, MappingTable);
										GenoCell{UResolGeno(k2)} = union(GenoCell{UResolGeno(k2)},tempIndex); 
									end
								end
								
								if(MaxIndex > 1100)
									break;
								end

							end %end of for loop k3
						end %end of else
					end %end of if these two genotypes are the same
					
					if(MaxIndex > 1100)
						break;
					end
				end %end of for loop k2
				
				if(MaxIndex > 1100)
						break;
				end
			end	%end of for loop k1
					
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
		
			disp(MaxIndex);
			if(MaxIndex > 1100)
					break;
			end
			
			%Write down trivial genotypes first
			tempVec = find(NeedResolve == 0);
			for i = 1: length(tempVec)
				fprintf('Genotype %d can be resolved directly!\n', UResolGeno(tempVec(i)));
				
				tempGeno = HapMat(UResolGeno(tempVec(i)),:);
				tempHap1 = strrep(tempGeno, '2', '0');
				tempHap2 = strrep(tempGeno, '2', '1');
		
				%Mark this genotype resolved
				ResolGeno(1, UResolGeno(tempVec(i))) = 1;
				
				NumOfUGenotype = NumOfUGenotype - 1;
				FinalSolution = union(FinalSolution, tempHap1, 'rows');
				FinalSolution = union(FinalSolution, tempHap2, 'rows');
				
				fprintf(Outfid, '%s = ', HapMat(UResolGeno(tempVec(i)),:));
				fprintf(Outfid, '%s + %s ;\n', tempHap1, tempHap2);			
			end
			
			%Create an array for the selection of each haplotype
			Selection = zeros(1, 1+MaxIndex);
			%Set x0 = 1;
			Selection(1,1) = 1;
			UResolGeno = find(ResolGeno == 0);
			ResolGeno2 = ResolGeno;
			
			%Emumerate each pair of possible haplotypes for all genotypes
			%Add all constraint of each genotype
			for i = 1:length(UResolGeno)
				%Record the index of non-2 element
				Not2 = [];
				for j = 1:NumOfSNP
					if(HapMat(UResolGeno(i),j) == '0' || HapMat(UResolGeno(i),j) == '1')
						Not2 = [Not2,j];
					end
				end %end of for loop j
		
				%Create the constraint matrix for this genotype
				Block1 = zeros(1+MaxIndex);
				Block2 = zeros(1, NumOfUGenotype);
		
				if(length(Not2) == NumOfSNP & NeedResolve(i) == 1)
					fprintf('Genotype %d does not need to resolve!\n',UResolGeno(i));
					
					Index = Find_Index(HapMat(UResolGeno(i),:), MappingTable);
					
					%this haplotype must be selected
					Block1(1, 1) = 1;
					Block1(1, Index+1) = 1;
					Block1(Index+1, 1) = 1;
					Block1(Index+1, Index+1) = 1;
		
					%Add the constraint for this genotype into the constraint matrix
					temp1cell = cell(2,1);
					temp1cell{1,1} = sparse(-Block1);
					temp1cell{2,1} = -Block2;
					ConstMat=[ConstMat temp1cell];
		
					%Mark the selection of this haplotype
					Selection(1, Index+1) = 1;
		
					%Mark this genotype resolved
					ResolGeno(1, UResolGeno(i)) = 1;
					%NumOfUGenotype = NumOfUGenotype - 1;
					FinalSolution = union(FinalSolution, HapMat(UResolGeno(i),:), 'rows');
					
					fprintf(Outfid, '%s = ', HapMat(UResolGeno(i),:));
					fprintf(Outfid, '%s + %s ;\n', HapMat(UResolGeno(i),:), HapMat(UResolGeno(i),:));
					
				else
					%unresolved genotypes
					fprintf('Genotype %d need to resolve!\n',UResolGeno(i));
		
					%Emumerate each pair of possible haplotypes for this genotype
					tempGenoCell = GenoCell{UResolGeno(i)};
					while(length(tempGenoCell) > 0)
						Index1 = tempGenoCell(1);
						p1 = MappingTable(tempGenoCell(1),:);
						p2 = Invert(p1, HapMat(UResolGeno(i),:));
						Index2 = Find_Index(p2, MappingTable);
						
						tempGenoCell = setdiff(tempGenoCell, Index1);
						tempGenoCell = setdiff(tempGenoCell, Index2);
						
						%Check if this haplotype is in the MappingTable
						if((Index1 ~= -1) & (Index2 ~= -1))
							Block1(1, 1) = Block1(1, 1)+1;
							Block1(1, Index1+1) = Block1(1, Index1+1)+0.5;
							Block1(1, Index2+1) = Block1(1, Index2+1)+0.5;
							Block1(Index1+1, 1) = Block1(Index1+1, 1)+0.5;
							Block1(Index2+1, 1) = Block1(Index2+1, 1)+0.5;
							Block1(Index1+1, Index2+1) = Block1(Index1+1, Index2+1)+0.5;
							Block1(Index2+1, Index1+1) = Block1(Index2+1, Index1+1)+0.5;
						else
							disp('Error\n');
						end
		
					end  %end of loop for all pairs of possible haplotypes
					
					%disp(Block1(1,1));
		
					%store the constraint matrix into ConstMat
					temp2cell = cell(2,1);
					temp2cell{1, 1} = sparse(-Block1);
					
					%store the constraint for >=4 slack variable
					%Block2(1, UResolGeno(i)) = -1;
					Block2(1, i) = -1;
					%disp(Block2);
					temp2cell{2,1} = -Block2;
					ConstMat = [ConstMat temp2cell];
		
				end %end of else for unresolved genotype
			end %end of loop for each genotype
		
			%Clear memory
			clear temp*;
			clear Block*;
		
			if(NumOfUGenotype == 0)
				%break the outer while loop
				break;
			end
			
			%Taget constraint: F0: x0+x1+x2...
			fprintf('Building the constraint matrix of objective function...\n');
			MainConst = eye(1+MaxIndex);
			MainConst(1:1+MaxIndex,1) = 1;
			MainConst(1,1:1+MaxIndex) = 1;
			MainConst(1,1) = MaxIndex;
			
			tempcell = cell(2,1);
			tempcell{1,1} = sparse(-MainConst);
			tempcell{2,1} = zeros(1, NumOfUGenotype);
			ConstMat = [tempcell ConstMat];
			clear MainConst;
			clear tempcell;
		
			%Build other Constraint: yii = 1
			fprintf('Building other constraint matrix...\n');
			tempMat = sparse(zeros(1+MaxIndex));
			tempVec = zeros(1, NumOfUGenotype);
			
			for m = 1:1+MaxIndex
				tempMat(m,m) = -1;
				if(m ~= 1)
					tempMat(m-1, m-1) = 0;
				end
				tempcell3 = cell(2,1);
				tempcell3{1,1} = tempMat;
				tempcell3{2,1} = tempVec;
				ConstMat = [ConstMat tempcell3];
			end
		
			clear temp*;
		
			%Set up initial parameters for the SDPA library
			mDIM = NumOfUGenotype+1+MaxIndex;
			nBLOCK = 2;
			BlockStruct = [1+MaxIndex, -NumOfUGenotype];
			ConstVector = [-4*ones(1, NumOfUGenotype) -ones(1, 1+MaxIndex)];
		
			%Invoke the SDPA library to solve the SDP problem
			fprintf('Solving SDPA...\n');
			[objVal,x,X,Y]=sdpam(mDIM,nBLOCK,BlockStruct,ConstVector,ConstMat);
			%disp(Y{1,1});
		
			%Perform the incomlete cholesky infinity factorization
			fprintf('Performing incomplete cholesky factorization...\n');
			Rinf = sparse(cholinc(sparse(Y{1,1}), 'inf'));
			%Rinf = sparse(chol(Y{1,1}));
			%disp(Rinf);
			
			clear X;
			clear Y;
			clear x;
			clear objVal;
			%clear mDIM;
			%clear BlockStruct;
			%clear ConstVector;
			
			%Randomized-Rounding phase
			fprintf('Performing randomized rounding...\n')
		
			%create two random unit vectors
			z1 = rand(1, 1+MaxIndex);
			z1 = z1/norm(z1, 2);
			z2 = rand(1, 1+MaxIndex);
			z2 = z2/norm(z2, 2);
			%disp(z1);
			%disp(z2);
		
			%Randomized rounding phase
			%Set xi = 1 or 0
			for n=2:1+MaxIndex
				if(Selection(1, n) == 0)
				%this haplotype is still unselected
					if((z1*Rinf(:,n))*(z1*Rinf(:,1)) > 0) & ((z2*Rinf(:,n))*(z2*Rinf(:,1)) > 0)
						Selection(1, n)=1;
					end
				end
			end
		
			%find the selected haplotype: 1s in the Selection vector
			tempVec = find(Selection);
			
			fprintf('\nResult of the %d-th iteration: \n', NumOfIteration);
			counter = 0;
			
			%check if any genotype still unresolved
			for o=1:NumOfGenotype
			
				%This genotype is unresolved; check if it can be resolved.
				if(ResolGeno2(1, o) == 0)
					%This counter is to count the unresolved genotypes
					counter = counter + 1;
					
					if(ResolGeno(1, o) == 0)
					
					%Enumerate each pair of selected haplotypes
					%Starting from 2 because Selection = 1+MaxIndex
					for l1=2:length(tempVec)-1
						for(l2=l1+1:length(tempVec))
							
							%Retrive the constraint matrix of the unresolved genotype by counter,
							%which will be used to check the resolution of that genotype
							%counter +1 because the first matrix in ConstMat is objective function
							tempMat = ConstMat{1, counter+1};
							
							%check the constraint matrix if any pair of haplotypes can resolve it
							%tempVec(1, l1) is the index of one haplotype
							if(tempMat(tempVec(1, l1), tempVec(1, l2)) ~= 0)
								%fprintf('%d and %d\n',tempVec(l1),tempVec(l2));
								%Mark this genotype resolved
								ResolGeno(1, o) = 1;
								%NumOfUGenotype = NumOfUGenotype - 1;
								
								fprintf('\nGenotype  %s resolved by\n', HapMat(o,:));
								p1 = MappingTable(tempVec(1, l1)-1,:);
								p2 = MappingTable(tempVec(1, l2)-1,:);
								
								fprintf('haplotype %s\n      and %s \n', p1, p2);
								
								FinalSolution = union(FinalSolution, [p1;p2], 'rows');
								
								fprintf(Outfid, '%s = ', HapMat(o,:));
								fprintf(Outfid, '%s + %s ;\n', p1, p2);
								break;
							end
						end %end of loop l2
						
						if(ResolGeno(1, o) == 1)
							break;
						end
					end %end of loop l1
				end %end of if
				end %end of if
			end
			%disp(Selection);
			%fprintf('\nResolved Genotypes: \n');
			
			clear temp*;
			clear ConstMat;
			clear Rinf;
			clear MappingTable;
			
			NumOfUGenotype = length(find(ResolGeno == 0));
			ResolGeno2 = ResolGeno;
		end %end of while loop
		
		if(MaxIndex > 1100)
			break;
		end
		
		fprintf('\n\n*********************************************\n');
		fprintf('******* SDPHapInfer Execution Complete ******\n');
		fprintf('*********************************************\n');
	
		[NumOfHaplotype, Column] = size(FinalSolution);
		fprintf('\nNumber of iterations: %d\n', NumOfIteration);
		fprintf('Number of selected haplotypes: %d\n', NumOfHaplotype);
		fprintf('List of selected haplotypes:\n');
		disp(FinalSolution);
		
		%Written the result into output file
		fprintf(Outfid, '\nNumber of iterations: %d\n', NumOfIteration);
		fprintf(Outfid, 'Number of selected haplotypes: %d\n', NumOfHaplotype);
		fprintf(Outfid, 'List of selected haplotypes:\n');
		for i=1:NumOfHaplotype
			fprintf(Outfid, '%s\n', FinalSolution(i, :));
		end
	
		fclose(Outfid);	
	
		%Copy tmp file to output file
		if(NumOfHaplotype < CurrentHaplotype)
			CurrentHaplotype = NumOfHaplotype;
			fprintf('Number of used haplotypes: %d\n', NumOfHaplotype);
			OutputFileName = [FileName '.out'];
			[Outfid2, message] = fopen(OutputFileName, 'w');
			[ftmp, message] = fopen('sdp.tmp','r');
			while (~feof(ftmp))
				line = fgetl(ftmp);
				fprintf(Outfid2, '%s\n', line);
			end
			fclose(ftmp);
			fclose(Outfid2);
		end
		
		%Stop timer	
		toc
	end	%end of else for opening file success
	
end	%end of iterations
